<?php
  /*
  ########################################################################################################
  convert incoming weather station data to WSWin CSV for WU and Ecowitt-Protocol
  save them into different wswin.csv files - one for each station
  only allowed stations are able to save data here
  only one record per minute allowed
  only records with newer time as last saved time allowed

  v0.1  09.04.2021 initial
  v0.1a 10.04.2021 initial reloaded
  v0.1b 19.08.2021 convert leaf wetness (WN35) from 0..99 to 0..15
                   replace all decimal points with comma for better Excel-handling ($useKomma)
  v0.2  28.01.2022 disable interval checking with $mininterval = 0;
  v0.2a 30.01.2022 alternative function to read the last line of the output file if tail command does not work (some hosters prevent this)

  Oliver Engel (olicat)
  https://www.wetterstationsforum.info/
  https://www.wxforum.net/

  script inspired by:
  Raffaello Di Martino http://www.kwos.it 2019 and the modifications made by Werner Krenn
  Thanks to Werner for WSWin
  ########################################################################################################
  */

  ########################################################################################################
  # adjust your settings below - critical are EW_allowed and/or WU_allowed
  ########################################################################################################
  # user have to authorize with PASSKEY (Ecowitt) or ID/PASSWORD (WU) - add your corresponding lines here
  # for Ecowitt stations pick the PASSKEY from the file configured under $errlog
  $EW_allowed = [
                  "000102030405060708090A0B0C0D0E0F",
                  "0102030405060708090A0B0C0D0E0F00",
                ];
  # for WU stations you have to configure a username and a corresponding password here
  $WU_allowed = array(
                  'user01' => 'pwduser01',
                  'user02' => 'pwduser02',
                );

  $allowAny = False;                                          # allow any weather station without user verification
  $mininterval = 60;                                          # accept only data if newer than n seconds; disable with 0
  $defTZ = 'Europe/Berlin';                                   # default time zone for local time
  $outdir = ".";                                              # destination dir for csv output
  $errlog = "error.log";                                      # file name for unknown weather stations
  $debug = False;                                             # for debugging purpose - switch to False
  $dbgfile = "debug.log";                                     # save all requests if $debug enabled
  $useKomma = True;                                           # use comma as decimal point

  ########################################################################################################
  # no need to make modifications below
  ########################################################################################################
  function ftoc($f) {                                          # convert Fahrenheit to Celsius
    if (isset($f) and is_numeric($f)) { $erg = round(($f - 32) * 5 / 9, 1); } else { $erg = ""; }
    return $erg;
  }

  function intohpa($f) {                                       # convert inch of mercury to hpa
    if (isset($f) and is_numeric($f)) { $erg = round($f * 33.86, 1); } else { $erg = ""; }
    return $erg;
  }

  function mphtokmh($f) {                                      # convert mph to kmh
    if (isset($f) and is_numeric($f)) { $erg = round($f * 1.60934, 2); } else { $erg = ""; }
    return $erg;
  }

  function intomm($f) {                                        # convert inch to mm
    if (isset($f) and is_numeric($f)) { $erg = round($f * 25.4, 2); } else { $erg = ""; }
    return $erg;
  }

  function leafTo15($f) {                                      # convert leaf wetness (WN35) from 0..99 to 0..15
    if (isset($f) and is_numeric($f)) { $erg = round($f / 6.6, 1); } else { $erg = ""; }
    return $erg;
  }

  function debugPrint($s) {                                    # in case of $debug = True print to web browser
    global $debug;
    if ($debug) { print("$s<br/>"); }
  }

  function WeatherTimeToWSWin($dateutc) {
    global $defTZ;
    $dateutc = str_replace("%20"," ",$dateutc);
    if (strlen($dateutc) !== 19 or substr($dateutc,4,1) !== "-" or substr($dateutc,7,1) !== "-" or substr($dateutc,13,1) !== ":" or substr($dateutc,16,1) !== ":") {
      date_default_timezone_set('UTC');
      $dateutc = date("Y-m-d+H:i:s",time());
    }
    date_default_timezone_set($defTZ);
    return date("d.m.Y;H:i", mktime(substr($dateutc,11,2),substr($dateutc,14,2),substr($dateutc,17,2),substr($dateutc,5,2),substr($dateutc,8,2),substr($dateutc,0,4))+date("Z"));
  }

  function WSWinTimeToTimeStamp($timestr) {
    # 09.04.2021;10:53
    if (strlen($timestr) !== 16 or substr($timestr,2,1) !== "." or substr($timestr,5,1) !== "." or substr($timestr,10,1) !== ";" or substr($timestr,13,1) !== ":") {
      $unixtime = time();
    } else {
      $unixtime = mktime(substr($timestr,11,2),substr($timestr,14,2),"00",substr($timestr,3,2),substr($timestr,0,2),substr($timestr,6,4));
    }
    return $unixtime;
  }

  function lastLine($fname) {                                  # read last line of a file
    $lastline = "";
    $fp = fopen($fname, "r");
    fseek($fp, filesize($fname)-512);                          # max. line length
    while (!feof($fp)) {
      $is = fgets($fp);
      if (!feof($fp)) { $lastline = trim($is); }
    }
    fclose($fp);
    return $lastline;
  }

  function checkData($a) {                                     # check auth and save unknown PASSKEY/user/password to a logfile, return WSWin time
    global $outdir, $allowAny, $mininterval, $isPOST, $EW_allowed, $WU_allowed, $errlog, $remote_ip;
    $givenPASSKEY = isset($a['PASSKEY']) ? $a['PASSKEY']: "";
    $user = isset($a['ID']) ? strtolower($a['ID']) : 'unknown user';
    $pass = isset($a['PASSWORD']) ? $a['PASSWORD'] : 'no password';
    $errcode = 0;
    # Abbruch wenn: PASSKEY nicht vorhanden und ID auch nicht
    if (!array_key_exists('PASSKEY', $a) and !isset($a['ID'])) {
      $errcode = 400;
      $errline = "bad request - no passkey nor ID sent";
      print("bad request");
    # Abbruch wenn: PASSKEY nicht bekannt und Passwort passt nicht zur ID und allowAny = False
    } elseif ((!in_array($givenPASSKEY,$EW_allowed,True) and $WU_allowed[$user] != $pass) and !$allowAny) {
      $errcode = 401;
      if ($isPOST) { $errline = $givenPASSKEY." is not allowed here"; } else { $errline = $user." is not allowed here or has submitted an incorrect password"; }
      print("unauthorized");
    # compare current time with last time - reject if current data is older than last known
    } else {
      $isttime = WeatherTimeToWSWin($a['dateutc']);
      $iststamp = WSWinTimeToTimeStamp(WeatherTimeToWSWin($a['dateutc']));
      if (isset($a['PASSKEY'])) { $outname = $a['PASSKEY']."-wswin.csv"; } else { $outname = $a['ID']."-wswin.csv"; }
      $outfile = $outdir."/".$outname;
      if (file_exists($outfile)) {                             # check file existence
        $file = escapeshellarg($outfile);                      # for security reasons
        $line = `tail -n 1 $outfile`;                          # last line in saved file
        if ($line == "") {                                     # in case tail is forbidden
          $line = lastLine($outfile);
        }
        $lasttime = substr($line,0,16);
        $laststamp = WSWinTimeToTimeStamp(substr($line,0,16));
        $diff = $iststamp - $laststamp;
        debugPrint("now: ".$isttime." (".$iststamp.") last: ".$lasttime." (".$laststamp.") diff: ".$diff."<br/>");
        if ($mininterval > 0 and $diff < $mininterval) {
          $errcode = 429;
          $errline = "too many requests - now: ".$isttime." (".$iststamp.") last: ".$lasttime." (".$laststamp.") diff: ".$diff;
          print("too many requests");
        }
      }
    }
    if ($errcode != 0) {
      $errline = date("d.m.Y")." ".date("H:i:s")." ".$remote_ip." ".$errcode." ".$errline."\r\n";
      if (isset($errlog)) { file_put_contents($errlog, $errline, FILE_APPEND); }
      http_response_code($errcode);
      exit($errcode);
    }
    return $isttime;
  }

  function getClientIPaddress(){
    foreach (array('HTTP_CLIENT_IP', 'HTTP_X_FORWARDED_FOR', 'HTTP_X_FORWARDED', 'HTTP_X_CLUSTER_CLIENT_IP', 'HTTP_FORWARDED_FOR', 'HTTP_FORWARDED', 'REMOTE_ADDR') as $key){
      if (array_key_exists($key, $_SERVER) === true){
        foreach (explode(',', $_SERVER[$key]) as $ip){
          return trim($ip);
        }
      }
    }
  }

  ########################################################################################################
  # main
  ########################################################################################################
  # some general vars
  $pd = $_POST;
  $gd = $_GET;
  $remote_ip = getClientIPaddress();
  if ($remote_ip == "") { $remote_ip = "from console"; }

  # header line for WSWin csv
  $hdr_csv = ";;1;17;133;2;18;35;36;45;134;42;41;3;19;4;20;5;21;6;22;7;23;8;24;29;30;31;32;25;26;27;28;13;14;15;16\r\n";

  # default array to fill with "" in case of absence
  $def_csv = ['tempinc','humidityin','baromrelhpa','tempc','humidity','windspeedkmh','winddir','windgustkmh','dailyrainmm','solarradiation','uv','temp1c','humidity1','temp2c','humidity2','temp3c','humidity3','temp4c','humidity4','temp5c','humidity5','temp6c','humidity6','soilmoisture1','soilmoisture2','soilmoisture3','soilmoisture4','leafwetness_ch1','leafwetness_ch2','leafwetness_ch3','leafwetness_ch4','tf_ch1c','tf_ch2c','tf_ch3c','tf_ch4c'];

  if (!empty($pd)) {
    $wd = $pd;                                                 # POST data available
    $isPOST = True;
  } else {
    $wd = $gd;                                                 # GET data available
    $isPOST = False;
  }

  if ($debug) {                                                # save all incoming traffic
    if ($isPOST) {
      $proto = "POST";
      $data = http_build_query($wd);
    } else {
      $proto = "GET ";
      $data = "";
    }
    if (isset($_SERVER["REQUEST_URI"])) { $uri = $_SERVER["REQUEST_URI"]; } else { $uri = ""; }
    $dbgline = date("d.m.Y")." ".date("H:i:s")." ".$remote_ip." ".$proto." ".$uri." ".$data."\r\n";
    file_put_contents($dbgfile, $dbgline, FILE_APPEND);
  }

  # check data for correct authentication and time
  $dateloc = checkData($wd);

  # for WU compatibility - fill fields with WU keys
  if (!$isPOST) {
    # only replace empty values
    if (!isset($wd['uv'])) {@$wd['uv'] = $wd['UV']; }
    if (!isset($wd['tempinf'])) {@$wd['tempinf'] = $wd['indoortempf']; }
    if (!isset($wd['humidityin'])) {@$wd['humidityin'] = $wd['indoorhumidity']; }
    if (!isset($wd['baromrelin'])) {@$wd['baromrelin'] = $wd['baromin']; }
    if (!isset($wd['soilmoisture1'])) {@$wd['soilmoisture1'] = $wd['soilmoisture']; }
    if (!isset($wd['tf_ch1'])) {@$wd['tf_ch1'] = $wd['soiltempf']; }
    if (!isset($wd['tf_ch2'])) {@$wd['tf_ch2'] = $wd['soiltemp2f']; }
    if (!isset($wd['tf_ch3'])) {@$wd['tf_ch3'] = $wd['soiltemp3f']; }
    if (!isset($wd['tf_ch4'])) {@$wd['tf_ch4'] = $wd['soiltemp4f']; }
    if (!isset($wd['leafwetness_ch1'])) {@$wd['leafwetness_ch1'] = $wd['leafwetness']; }
    if (!isset($wd['leafwetness_ch2'])) {@$wd['leafwetness_ch2'] = $wd['leafwetness2']; }
    if (!isset($wd['leafwetness_ch3'])) {@$wd['leafwetness_ch3'] = $wd['leafwetness3']; }
    if (!isset($wd['leafwetness_ch4'])) {@$wd['leafwetness_ch4'] = $wd['leafwetness4']; }
  }

  # convert values to metric system as WSWin expect this
  @$wd['tempinc'] = ftoc($wd['tempinf']);
  @$wd['tempc'] = ftoc($wd['tempf']);
  for ($i = 1; $i <= 8; $i++) { @$wd["temp".$i."c"] = ftoc($wd["temp".$i."f"]); }
  for ($i = 1; $i <= 4; $i++) { @$wd["tf_ch".$i."c"] = ftoc($wd["tf_ch".$i]); }
  @$wd['baromrelhpa'] = intohpa($wd['baromrelin']);
  @$wd['windspeedkmh'] = mphtokmh($wd['windspeedmph']);
  @$wd['windgustkmh'] = mphtokmh($wd['windgustmph']);
  @$wd['dailyrainmm'] = intomm($wd['dailyrainin']);
  
  # convert leaf wetness (WN35) from 0..99 to 0..15
  for ($i = 1; $i <= 4; $i++) { @$wd["leafwetness_ch".$i] = leafTo15($wd["leafwetness_ch".$i]); }

  # create csv output line
  $csvline = "$dateloc";                                       # WSWin expects local time in format DD.MM.YYYY;hh:mm
  foreach ($def_csv as $element) {
    if (!array_key_exists($element, $wd)) {
      $wd[$element] = "";
      debugPrint($element.": does not exist!");                # show empty values
    } else { debugPrint($element.": ".$wd[$element]); }
    # add element to CSV line and replace any "." with "," if $useKomma is set
    if ($useKomma) { $csvline = $csvline.";".str_replace('.',',',strval($wd[$element])); } else { $csvline = $csvline.";".$wd[$element]; }
  }

  # create file with auto naming
  if (isset($wd['PASSKEY'])) { $outname = $wd['PASSKEY']."-wswin.csv"; } else { $outname = $wd['ID']."-wswin.csv"; }
  $outfile = $outdir."/".$outname;
  if (!file_exists($outfile)) {                                # check file existence; write csv header while creating a new file
    file_put_contents($outfile, $hdr_csv, FILE_APPEND);
  }

  # add line to csv
  $csvline = $csvline."\r\n";
  $wrote = file_put_contents($outfile, $csvline, FILE_APPEND);
  debugPrint("<br/>CSV line:");
  debugPrint($csvline."<br/>");

  # check if appending data worked 
  debugPrint("written: ".$wrote);
  if ($wrote > 0) {
    $errcode = 201;
    print("success\n");
  } else { 
    $errcode = 500; 
    print("server error\n");
  }
  
  # send response
  http_response_code($errcode);
?>

